"""
    AVM SmartHome Actor
    ~~~~~~~~~~~~~~~~~~~
"""

import logging
logger = logging.getLogger(__name__)

class Actor(object):
    """
    Represents a single SmartHome actor.
    You usally don't create that class yourself, use FritzBox.get_actors
    instead.
    """

    def __init__(self, fritzbox, device):
        self.box = fritzbox

        self.actor_id = device.attrib['identifier']
        self.device_id = device.attrib['id']
        self.name = device.find('name').text
        self.fwversion = device.attrib['fwversion']
        self.productname = device.attrib['productname']
        self.manufacturer = device.attrib['manufacturer']
        self.functionbitmask = int(device.attrib['functionbitmask'])

        self.has_powermeter = self.functionbitmask & (1 << 7) > 0
        self.has_temperature = self.functionbitmask & (1 << 8) > 0
        self.has_switch = self.functionbitmask & (1 << 9) > 0

        self.temperature = 0.0
        if self.has_temperature:
            if device.find("temperature").find("celsius").text is not None:
                self.temperature = int(device.find("temperature").find("celsius").text) / 10
            else:
                logger.warn("Actor " + self.name + " seems offline. Returning None as temperature.")
                self.temperature=None

    def switch_on(self):
        """
        Set the power switch to ON.
        """
        return self.box.set_switch_on(self.actor_id)

    def switch_off(self):
        """
        Set the power switch to OFF.
        """
        return self.box.set_switch_off(self.actor_id)

    def get_state(self):
        """
        Get the current switch state.
        """
        return bool(
            int(self.box.homeautoswitch("getswitchstate", self.actor_id))
        )

    def get_present(self):
        """
        Check if the registered actor is currently present (reachable).
        """
        return bool(
            int(self.box.homeautoswitch("getswitchpresent", self.actor_id))
        )

    def get_power(self):
        """
        Returns the current power usage in milliWatts.
        Attention: Returns None if the value can't be queried or is unknown.
        """
        value = self.box.homeautoswitch("getswitchpower", self.actor_id)
        return int(value) if value.isdigit() else None

    def get_energy(self):
        """
        Returns the consumed energy since the start of the statistics in Wh.
        Attention: Returns None if the value can't be queried or is unknown.
        """
        value = self.box.homeautoswitch("getswitchenergy", self.actor_id)
        return int(value) if value.isdigit() else None

    def get_temperature(self):
        """
        Returns the current environment temperature.
        Attention: Returns None if the value can't be queried or is unknown.
        """
        #raise NotImplementedError("This should work according to the AVM docs, but don't...")
        value = self.box.homeautoswitch("gettemperature", self.actor_id)
        if value.isdigit():
            return float(value)/10
        else:
            return None

    def get_consumption(self, timerange="10"):
        """
        Return the energy report for the device.
        """
        return self.box.get_consumption(self.device_id, timerange)

    def __repr__(self):
        return u"<Actor {}>".format(self.name)



"""
    AVM Fritz!BOX SmartHome Client
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Dokumentation zum Login-Verfahren:
    http://www.avm.de/de/Extern/files/session_id/AVM_Technical_Note_-_Session_ID.pdf

    Smart Home Interface:
    http://www.avm.de/de/Extern/files/session_id/AHA-HTTP-Interface.pdf
"""

from __future__ import print_function, division

import hashlib
from collections import namedtuple
from xml.etree import ElementTree as ET

from requests import Session
try:
    from bs4 import BeautifulSoup
except ImportError:
    BeautifulSoup = None

from .actor import Actor


Device = namedtuple("Device", "deviceid connectstate switchstate")
LogEntry = namedtuple("LogEntry", "date time message hash")


class FritzBox(object):
    """
    Provides easy access to a FritzBOX's SmartHome functions,
    which are poorly documented by AVM...

    A note about SIDs:
     They expire after some time. If you have a long-running daemon,
     you should call login() every 10 minutes or so else you'll get
     nice 403 errors.
    """

    def __init__(self, ip, username, password, use_tls=False):
        if use_tls:
            self.base_url = 'https://' + ip
        else:
            self.base_url = 'http://' + ip
        self.username = username
        self.password = password
        self.sid = None

        self.session = Session()

    def login(self):
        """
        Try to login and set the internal session id.

        Please note:
        - Any failed login resets all existing session ids, even of
          other users.
        - SIDs expire after some time
        """
        response = self.session.get(self.base_url + '/login_sid.lua', timeout=10)
        xml = ET.fromstring(response.text)
        if xml.find('SID').text == "0000000000000000":
            challenge = xml.find('Challenge').text
            url = self.base_url + "/login_sid.lua"
            response = self.session.get(url, params={
                "username": self.username,
                "response": self.calculate_response(challenge, self.password),
            }, timeout=10)
            xml = ET.fromstring(response.text)
            sid = xml.find('SID').text
            if xml.find('SID').text == "0000000000000000":
                blocktime = int(xml.find('BlockTime').text)
                exc = Exception("Login failed, please wait {} seconds".format(
                    blocktime
                ))
                exc.blocktime = blocktime
                raise exc
            self.sid = sid
            return sid

    def calculate_response(self, challenge, password):
        """Calculate response for the challenge-response authentication"""
        to_hash = (challenge + "-" + password).encode("UTF-16LE")
        hashed = hashlib.md5(to_hash).hexdigest()
        return "{0}-{1}".format(challenge, hashed)

    #
    # Useful public methods
    #
    def get_actors(self):
        """
        Returns a list of Actor objects for querying SmartHome devices.

        This is currently the only working method for getting temperature data.
        """
        devices = self.homeautoswitch("getdevicelistinfos")
        xml = ET.fromstring(devices)

        actors = []
        for device in xml.findall('device'):
            actors.append(Actor(fritzbox=self, device=device))

        return actors

    def get_actor_by_ain(self, ain):
        """
        Return a actor identified by it's ain or return None
        """
        for actor in self.get_actors():
            if actor.actor_id == ain:
                return actor

    #
    # "Private" methods
    #
    def homeautoswitch(self, cmd, ain=None):
        """
        Call a switch method.
        Should only be used by internal library functions.
        """
        assert self.sid, "Not logged in"
        params = {
            'switchcmd': cmd,
            'sid': self.sid,
        }
        if ain:
            params['ain'] = ain
        url = self.base_url + '/webservices/homeautoswitch.lua'
        response = self.session.get(url, params=params, timeout=10)
        response.raise_for_status()
        return response.text.strip()

    def get_switch_actors(self):
        """
        Get information about all actors

        This needs 1+(5n) requests where n = number of actors registered

        Deprecated, use get_actors instead.

        Returns a dict:
        [ain] = {
            'name': Name of actor,
            'state': Powerstate (boolean)
            'present': Connected to server? (boolean)
            'power': Current power consumption in mW
            'energy': Used energy in Wh since last energy reset
            'temperature': Current environment temperature in celsius
        }
        """
        actors = {}
        for ain in self.homeautoswitch("getswitchlist").split(','):
            actors[ain] = {
                'name': self.homeautoswitch("getswitchname", ain),
                'state': bool(self.homeautoswitch("getswitchstate", ain)),
                'present': bool(self.homeautoswitch("getswitchpresent", ain)),
                'power': self.homeautoswitch("getswitchpower", ain),
                'energy': self.homeautoswitch("getswitchenergy", ain),
                'temperature': self.homeautoswitch("getswitchtemperature", ain),
            }
        return actors

    def set_switch_on(self, ain):
        """Switch the power of a actor ON"""
        return self.homeautoswitch('setswitchon', ain)

    def set_switch_off(self, ain):
        """Switch the power of a actor OFF"""
        return self.homeautoswitch('setswitchoff', ain)

    def set_switch_toggle(self, ain):
        """Toggle a power switch and return the new state"""
        return self.homeautoswitch('setswitchtoggle', ain)

    #
    # DeviceID based methods
    #
    # Inspired by:
    # https://github.com/valpo/fritzbox/blob/master/fritzbox/fritzautohome.py
    #

    def get_devices(self):
        """
        Return a list of devices.
        Deprecated, use get_actors instead.
        """
        url = self.base_url + '/net/home_auto_query.lua'
        response = self.session.get(url, params={
            'sid': self.sid,
            'command': 'AllOutletStates',
            'xhr': 0,
        }, timeout=15)
        response.raise_for_status()
        data = response.json()
        count = int(data["Outlet_count"])
        devices = []
        for i in range(1, count + 1):
            device = Device(
                int(data["DeviceID_{0}".format(i)]),
                int(data["DeviceConnectState_{0}".format(i)]),
                int(data["DeviceSwitchState_{0}".format(i)])
            )
            devices.append(device)
        return devices

    def get_consumption(self, deviceid, timerange="10"):
        """
        Return all available energy consumption data for the device.
        You need to divice watt_values by 100 and volt_values by 1000
        to get the "real" values.

        :return: dict
        """
        tranges = ("10", "24h", "month", "year")
        if timerange not in tranges:
            raise ValueError(
                "Unknown timerange. Possible values are: {0}".format(tranges)
            )

        url = self.base_url + "/net/home_auto_query.lua"
        response = self.session.get(url, params={
            'sid': self.sid,
            'command': 'EnergyStats_{0}'.format(timerange),
            'id': deviceid,
            'xhr': 0,
        }, timeout=15)
        response.raise_for_status()

        data = response.json()
        result = {}

        # Single result values
        values_map = {
            'MM_Value_Amp': 'mm_value_amp',
            'MM_Value_Power': 'mm_value_power',
            'MM_Value_Volt': 'mm_value_volt',

            'EnStats_average_value': 'enstats_average_value',
            'EnStats_max_value': 'enstats_max_value',
            'EnStats_min_value': 'enstats_min_value',
            'EnStats_timer_type': 'enstats_timer_type',

            'sum_Day': 'sum_day',
            'sum_Month': 'sum_month',
            'sum_Year': 'sum_year',
        }
        for avm_key, py_key in values_map.items():
            result[py_key] = int(data[avm_key])

        # Stats counts
        count = int(data["EnStats_count"])
        watt_values = [None for i in range(count)]
        volt_values = [None for i in range(count)]
        for i in range(1, count + 1):
            watt_values[i - 1] = int(data["EnStats_watt_value_{}".format(i)])
            volt_values[i - 1] = int(data["EnStats_volt_value_{}".format(i)])

        result['watt_values'] = watt_values
        result['volt_values'] = volt_values

        return result

    def get_logs(self):
        """
        Return the system logs since the last reboot.
        """
        assert BeautifulSoup, "Please install bs4 to use this method"

        url = self.base_url + "/system/syslog.lua"
        response = self.session.get(url, params={
            'sid': self.sid,
            'stylemode': 'print',
        }, timeout=15)
        response.raise_for_status()

        entries = []
        tree = BeautifulSoup(response.text)
        rows = tree.find('table').find_all('tr')
        for row in rows:
            columns = row.find_all("td")
            date = columns[0].string
            time = columns[1].string
            message = columns[2].find("a").string

            merged = "{} {} {}".format(date, time, message.encode("UTF-8"))
            msg_hash = hashlib.md5(merged).hexdigest()
            entries.append(LogEntry(date, time, message, msg_hash))
        return entries



# coding: utf-8
"""
    FRITZ!Box SmartHome Client
    ~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


from __future__ import print_function, division

import re
import time
import json
import socket

import click

from .fritz import FritzBox


@click.group()
@click.option('--host', default='169.254.1.1') # fritzbox "emergency" IP
@click.option('--username', default='smarthome')
@click.option('--password', default='smarthome')
@click.pass_context
def cli(context, host, username, password):
    """
    FritzBox SmartHome Tool

    \b
    Provides the following functions:
    - A easy to use library for querying SmartHome actors
    - This CLI tool for testing
    - A carbon client for pipeing data into graphite
    """
    context.obj = FritzBox(host, username, password)


@cli.command()
@click.pass_context
def actors(context):
    """Display a list of actors"""
    fritz = context.obj
    fritz.login()

    for actor in fritz.get_actors():
        click.echo("{} ({} {}; AIN {})".format(
            actor.name,
            actor.manufacturer,
            actor.productname,
            actor.actor_id,
        ))


@cli.command()
@click.option('--features', type=bool, default=False, help="Show device features")
@click.pass_context
def energy(context, features):
    """Display energy stats of all actors"""
    fritz = context.obj
    fritz.login()

    for actor in fritz.get_actors():
        click.echo("{} ({}): {:.2f} Watt current, {:.3f} wH total, {:.2f} °C".format(
            actor.name,
            actor.actor_id,
            (actor.get_power() or 0.0) / 1000,
            (actor.get_energy() or 0.0) / 100,
            actor.temperature
        ))
        if features:
            click.echo("  Features: PowerMeter: {}, Temperatur: {}, Switch: {}".format(
                actor.has_powermeter, actor.has_temperature, actor.has_switch
            ))


@cli.command()
@click.argument('server')
@click.option('--port', type=int, default=2003)
@click.option('--interval', type=int, default=10)
@click.option('--prefix', default="smarthome")
@click.pass_context
def graphite(context, server, port, interval, prefix):
    """Display energy stats of all actors"""
    fritz = context.obj
    fritz.login()
    sid_ttl = time.time() + 600

    # Find actors and create carbon keys
    click.echo(" * Requesting actors list")
    simple_chars = re.compile('[^A-Za-z0-9]+')
    actors = fritz.get_actors()
    keys = {}
    for actor in actors:
        keys[actor.name] = "{}.{}".format(
            prefix,
            simple_chars.sub('_', actor.name)
        )

    # Connect to carbon
    click.echo(" * Trying to connect to carbon")
    timeout = 2
    sock = socket.socket()
    sock.settimeout(timeout)
    try:
        sock.connect((server, port))
    except socket.timeout:
        raise Exception("Took over {} second(s) to connect to {}".format(
            timeout, server
        ))
    except Exception as error:
        raise Exception("unknown exception while connecting to {} - {}".format(
            server, error
        ))

    def send(key, value):
        """Send a key-value-pair to carbon"""
        now = int(time.time())
        payload = "{} {} {}\n".format(key, value, now)
        sock.sendall(payload)

    while True:
        if time.time() > sid_ttl:
            click.echo(" * Requesting new SID")
            fritz.login()
            sid_ttl = time.time() + 600

        click.echo(" * Requesting statistics")
        for actor in actors:
            power = actor.get_power()
            total = actor.get_energy()
            click.echo("   -> {}: {:.2f} Watt current, {:.3f} wH total".format(
                actor.name, power / 1000, total / 100
            ))

            send(keys[actor.name] + '.current', power)
            send(keys[actor.name] + '.total', total)

        time.sleep(interval)


@cli.command(name="switch-on")
@click.argument('ain')
@click.pass_context
def switch_on(context, ain):
    """Switch an actor's power to ON"""
    context.obj.login()
    actor = context.obj.get_actor_by_ain(ain)
    if actor:
        click.echo("Switching {} on".format(actor.name))
        actor.switch_on()
    else:
        click.echo("Actor not found: {}".format(ain))


@cli.command(name="switch-off")
@click.argument('ain')
@click.pass_context
def switch_off(context, ain):
    """Switch an actor's power to OFF"""
    context.obj.login()
    actor = context.obj.get_actor_by_ain(ain)
    if actor:
        click.echo("Switching {} off".format(actor.name))
        actor.switch_off()
    else:
        click.echo("Actor not found: {}".format(ain))


@cli.command(name="switch-state")
@click.argument('ain')
@click.pass_context
def switch_state(context, ain):
    """Get an actor's power state"""
    context.obj.login()
    actor = context.obj.get_actor_by_ain(ain)
    if actor:
        click.echo("State for {} is: {}".format(ain,'ON' if actor.get_state() else 'OFF'))
    else:
        click.echo("Actor not found: {}".format(ain))


@cli.command(name="switch-toggle")
@click.argument('ain')
@click.pass_context
def switch_toggle(context, ain):
    """Toggle an actor's power state"""
    context.obj.login()
    actor = context.obj.get_actor_by_ain(ain)
    if actor:
        if actor.get_state():
            actor.switch_off()
            click.echo("State for {} is now OFF".format(ain))
        else:
            actor.switch_on()
            click.echo("State for {} is now ON".format(ain))
    else:
        click.echo("Actor not found: {}".format(ain))


@cli.command()
@click.option('--format', type=click.Choice(['plain', 'json']),
              default='plain')
@click.pass_context
def logs(context, format):
    """Show system logs since last reboot"""
    fritz = context.obj
    fritz.login()

    messages = fritz.get_logs()
    if format == "plain":
        for msg in messages:
            merged = "{} {} {}".format(msg.date, msg.time, msg.message.encode("UTF-8"))
            click.echo(merged)

    if format == "json":
        entries = [msg._asdict() for msg in messages]
        click.echo(json.dumps({
            "entries": entries,
        }))


if __name__ == '__main__':
    cli()




from setuptools import setup, find_packages

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

try:
    with open(path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = f.read()
except FileNotFoundError:
    long_description = "(README.md not found)"

setup(
    name="fritzhome",
    version="1.0.3",

    description="Query information from your FRITZ!Box (mostly energy)",
    long_description=long_description,

    url="https://github.com/DerMitch/fritzbox-smarthome",

    author="Michael Mayr",
    author_email="michael@dermitch.de",

    license="MIT",

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],

    keywords="fritzbox smarthome avm energy",

    packages=["fritzhome"],

    install_requires=[
        'requests>=2.12.0',
        'click>=6.0.0',
    ],

    entry_points={
        'console_scripts': [
            'fritzhome=fritzhome.__main__:cli',
        ],
    }
)
